//==================================
// PEDUMP - Matt Pietrek 1997
// FILE: LIBDUMP.C
//==================================

#include "Common.h"
#include "ExtrnVar.h"

static const char * PszLongnames = NULL;

static DWORD ConvertBigEndian (DWORD bigEndian);

static void DisplayArchiveMemberHeader (const IMAGE_ARCHIVE_MEMBER_HEADER *pArchHeader,
                                        DWORD fileOffset)
{
    printf("Archive Member Header (%08lX):\n", fileOffset);

    printf("  Name:     %.16s", pArchHeader->Name);
    if ( pArchHeader->Name[0] == '/' && isdigit(pArchHeader->Name[1]) )
        printf( "  (%s)\n", PszLongnames + atoi((char *)pArchHeader->Name+1) );
    printf("\n");

    char szDateAsLong[64];
    sprintf (szDateAsLong, "%.12s", pArchHeader->Date);
    time_t dateAsLong = atol(szDateAsLong);

    printf("  Date:     %.12s %s", pArchHeader->Date, ctime(&dateAsLong));
    printf("  UserID:   %.6s\n",   pArchHeader->UserID);
    printf("  GroupID:  %.6s\n",   pArchHeader->GroupID);
    printf("  Mode:     %.8s\n",   pArchHeader->Mode);
    printf("  Size:     %.10s\n",  pArchHeader->Size);
}

static void DumpFirstLinkerMember (const void *p)
{
    DWORD        cSymbols = *(const DWORD*)p;
    unsigned     i;
    const DWORD *pMemberOffsets = MakePtr (const DWORD *, p, 4);
    const char  *pSymbolName;

    cSymbols    = ConvertBigEndian (cSymbols);
    pSymbolName = MakePtr (const char *, pMemberOffsets, 4 * cSymbols);

    printf ("First Linker Member:\n");
    printf ( "  Symbols:         %08lX\n", cSymbols );
    printf ( "  MbrOffs   Name\n  --------  ----\n" );

    for ( i = 0; i < cSymbols; i++ )
    {
        DWORD offset = ConvertBigEndian (*pMemberOffsets);

        printf ("  %08lX  %s\n", offset, undecorate_name(pSymbolName));
        pMemberOffsets++;
        pSymbolName += strlen(pSymbolName) + 1;
    }
}

static void DumpSecondLinkerMember (const void *p)
{
    DWORD        i, cArchiveMembers = *(const DWORD*)p;
    const DWORD *pMemberOffsets = MakePtr (const DWORD *, p, sizeof(DWORD));
    DWORD        cSymbols;
    const char  *pSymbolName;
    const WORD  *pIndices;

//  cArchiveMembers = cArchiveMembers;

    // The number of symbols is in the DWORD right past the end of the
    // member offset array.
    cSymbols = pMemberOffsets[cArchiveMembers];

    pIndices    = MakePtr (const WORD *, p, 4 + cArchiveMembers * sizeof(DWORD) + 4);
    pSymbolName = MakePtr (const char *, pIndices, cSymbols * sizeof(WORD));

    printf ("Second Linker Member:\n");
    printf ( "  Archive Members: %08lX\n", cArchiveMembers );
    printf ( "  Symbols:         %08lX\n", cSymbols );
    printf ( "  MbrOffs   Name\n  --------  ----\n" );

    for ( i = 0; i < cSymbols; i++ )
    {
        printf("  %08lX  %s\n", pMemberOffsets [pIndices[i] - 1], undecorate_name(pSymbolName));
        pSymbolName += strlen(pSymbolName) + 1;
    }
}

static void DumpLongnamesMember (const void *p, DWORD len)
{
    const char *pszName = (const char*) p;
    DWORD       offset = 0;

    PszLongnames = (const char *)p;  // Save off pointer for use when dumping
                                     // out OBJ member names

    printf ("Longnames:\n");

    // The longnames member is a series of null-terminated string.  Print
    // out the offset of each string (in decimal), followed by the string.
    while (offset < len)
    {
        unsigned cbString = lstrlen (pszName) + 1;

        printf ("  %05lu: %s\n", offset, pszName);
        offset  += cbString;
        pszName += cbString;
    }
}

void DumpLibFile (const void *lpFileBase)
{
    const IMAGE_ARCHIVE_MEMBER_HEADER *pArchHeader;
    BOOL  fSawFirstLinkerMember = FALSE;
    BOOL  fSawSecondLinkerMember = FALSE;

    if (strncmp((const char*)lpFileBase, IMAGE_ARCHIVE_START, IMAGE_ARCHIVE_START_SIZE))
    {
        printf ("Not a valid .LIB file - signature not found\n");
        return;
    }

    pArchHeader = MakePtr (const IMAGE_ARCHIVE_MEMBER_HEADER *, lpFileBase,
                           IMAGE_ARCHIVE_START_SIZE);

    while (pArchHeader)
    {
        DWORD thisMemberSize;
        char  name [sizeof(pArchHeader->Name)+1];

        DisplayArchiveMemberHeader (pArchHeader,
                                    (BYTE*)pArchHeader - (BYTE*)lpFileBase);
        strncpy (name, (const char*)pArchHeader->Name, sizeof(name));
        name [sizeof(pArchHeader->Name)] = '\0';

        printf ("\n");

        if (!strncmp(name, IMAGE_ARCHIVE_LINKER_MEMBER, 16))
        {
            if (!fSawFirstLinkerMember)
            {
                DumpFirstLinkerMember ((const void*)(pArchHeader + 1));
                printf ("\n");
                fSawFirstLinkerMember = TRUE;
            }
            else if (!fSawSecondLinkerMember)
            {
                DumpSecondLinkerMember ((const void*)(pArchHeader + 1));
                printf ("\n");
                fSawSecondLinkerMember = TRUE;
            }
        }
        else if (!strncmp(name, IMAGE_ARCHIVE_LONGNAMES_MEMBER, 16))
        {
            DumpLongnamesMember ((const void *)(pArchHeader + 1),
                                 atoi((char*)pArchHeader->Size) );
            printf ("\n");
        }
        else if (strstr(name,".dll/"))
        {
            // DumpExeFile( (PIMAGE_DOS_HEADER)(pArchHeader + 1) );
        }
        else    // It's an OBJ file
        {
            DumpObjFile ((const IMAGE_FILE_HEADER*)(pArchHeader + 1));
        }

        // Calculate how big this member is (it's originally stored as
        // as ASCII string.
        thisMemberSize = atoi ((char *)pArchHeader->Size) +
                         IMAGE_SIZEOF_ARCHIVE_MEMBER_HDR;

        thisMemberSize = (thisMemberSize+1) & ~1;   // Round up

        // Get a pointer to the next archive member
        pArchHeader = MakePtr (const IMAGE_ARCHIVE_MEMBER_HEADER *,
                               pArchHeader, thisMemberSize);

        if (!PTR_IN_RANGE (&pArchHeader->EndHeader,2))
           break;

        // Bail out if we don't see the EndHeader signature in the next record
        if (strncmp((const char*)pArchHeader->EndHeader, IMAGE_ARCHIVE_END, 2))
            break;
    }
}

// Routine to convert from big endian to little endian
static DWORD ConvertBigEndian (DWORD bigEndian)
{
    DWORD temp = 0;

    // printf( "bigEndian: %08X\n", bigEndian );

    temp |= bigEndian >> 24;
    temp |= ((bigEndian & 0x00FF0000) >> 8);
    temp |= ((bigEndian & 0x0000FF00) << 8);
    temp |= ((bigEndian & 0x000000FF) << 24);

    return temp;
}

