//==================================
// PEDUMP - Matt Pietrek 1994-1998
// FILE: PEDUMP.C
//==================================

/* Some good refs on PE:
 *   http://www.delphibasics.info/home/delphibasicsarticles/anin-depthlookintothewin32portableexecutablefileformat-part1
 *   http://www.delphibasics.info/home/delphibasicsarticles/anin-depthlookintothewin32portableexecutablefileformat-part2
 */

#include "Common.h"
#include "ExtrnVar.h"
#include "getopt.h"

#include <signal.h>
#include <setjmp.h>

#undef STANDALONE_DEMANGLER

/*
 * Compile with the latest GCC version of 'mangle.c' downloaded
 * from the mirror at: https://github.com/gcc-mirror/gcc/blob/master/gcc/cp/mangle.c
 */
#if defined(USE_GCC_MANGLE_LATEST) && (USE_GCC_MANGLE_LATEST == 1)
  #include "gcc/latest-mangle.cc"

#elif defined(USE_CP_DEMANGLE2)
  #include "gcc/cp-demangle2.c"

#else
  #include "gcc/cp-demangle.c"
  #include "gcc/dyn-string.c"
#endif

char *program_name;   /* for getopt.c */

#if !defined(__MINGW64_VERSION_MAJOR)
  /*
   * MinGW's CRT globs the command line by default.
   * So this shouldn't hurt.
   */
  int _CRT_glob = 2;     /* == __CRT_GLOB_USE_MINGW__ */

  /* MinGW-64's CRT seems to NOT glob the cmd-line by default.
   */
  int _dowildcard = 1;

#else
  int _dowildcard = -1;
#endif

/* For MSVC we MUST link with the 'setargv.obj' file to get
 * cmd-line globbing.
 */

// Global variables set here, and used in EXEDUMP.C and OBJDUMP.C
BOOL fShowRelocations = FALSE;
BOOL fShowRawSectionData = FALSE;
BOOL fShowSymbolTable = FALSE;
BOOL fShowLineNumbers = FALSE;
BOOL fShowIATentries = FALSE;
BOOL fShowPDATA = FALSE;
BOOL fShowResources = FALSE;
BOOL fDemangle = TRUE;
int  fDebug = 0;

ULONGLONG     g_mmap_start = 0ULL;
ULONGLONG     g_mmap_end = 0ULL;
LARGE_INTEGER g_file_size;

func_UnDecorateSymbolName p_UnDecorateSymbolName = NULL;
func_CheckSumMappedFile   p_CheckSumMappedFile   = NULL;

typedef void (*sig_handler_t) (int);

static sig_handler_t sigsegv_handler_old;
static jmp_buf       sigsegv_buf;

static const char HelpText[] =
  "PEDUMP - Win32/COFF EXE/OBJ/LIB file dumper - 1998 Matt Pietrek\n\n"
  "Syntax: PEDUMP [switches] filename(s)\n\n"
  "  -A    include everything in dump.\n"
  "  -B    show base relocations.\n"
  "  -D    do not demangle C++ names.\n"
  "  -H    include hex dump of sections.\n"
  "  -I    include Import Address Table thunk addresses.\n"
  "  -L    include line number information.\n"
  "  -P    include PDATA (runtime functions).\n"
  "  -R    include detailed resources (stringtables and dialogs).\n"
  "  -S    show symbol table.\n"
  "  -d    increase debug-level.\n";

static void sigsegv_handler (int sig)
{
  if (sig == SIGSEGV)
     longjmp (sigsegv_buf,1);
}

static void init_dbghelp (void)
{
  HINSTANCE mod = LoadLibrary ("dbghelp.dll");

  if (mod && mod != INVALID_HANDLE_VALUE)
     p_UnDecorateSymbolName = (func_UnDecorateSymbolName) GetProcAddress (mod, "UnDecorateSymbolName");
}

static void init_imagehlp (void)
{
  HINSTANCE mod = LoadLibrary ("imagehlp.dll");

  if (mod && mod != INVALID_HANDLE_VALUE)
     p_CheckSumMappedFile = (func_CheckSumMappedFile) GetProcAddress (mod, "CheckSumMappedFile");
}


//
// Open up a file, memory map it, and call the appropriate dumping routine
//
static void DumpFile (const char *filename)
{
    const IMAGE_DOS_HEADER  *dosHeader;
    const IMAGE_FILE_HEADER *pImgFileHdr;
    HANDLE                   hFile, hFileMapping;
    void                    *lpFileBase;

    hFile = CreateFile (filename, GENERIC_READ, FILE_SHARE_READ, NULL,
                        OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);

    if (hFile == INVALID_HANDLE_VALUE)
    {
        printf("Couldn't open \"%s\" with CreateFile()\n", filename);
        return;
    }

    memset (&g_file_size, 0, sizeof(g_file_size));
    GetFileSizeEx (hFile, &g_file_size);

    hFileMapping = CreateFileMapping (hFile, NULL, PAGE_READONLY, 0, 0, NULL);
    if (hFileMapping == 0)
    {
        CloseHandle (hFile);
        printf ("Couldn't open file mapping with CreateFileMapping()\n");
        return;
    }

    lpFileBase = MapViewOfFile (hFileMapping, FILE_MAP_READ, 0, 0, 0);
    if (!lpFileBase)
    {
        CloseHandle (hFileMapping);
        CloseHandle (hFile);
        printf ("Couldn't map view of file with MapViewOfFile()\n");
        return;
    }

    g_mmap_start = (ULONGLONG) lpFileBase;
    g_mmap_end   = (ULONGLONG) lpFileBase + g_file_size.QuadPart;

    printf ("Dump of file %s (%I64u bytes).\n\n", filename, g_file_size.QuadPart);

    dosHeader   = (const IMAGE_DOS_HEADER*) lpFileBase;
    pImgFileHdr = (const IMAGE_FILE_HEADER*) lpFileBase;

    if (dosHeader->e_magic == IMAGE_DOS_SIGNATURE)
    {
        DumpExeFile (dosHeader);
    }
    else if (dosHeader->e_magic == IMAGE_SEPARATE_DEBUG_SIGNATURE)
    {
        DumpDbgFile ((PIMAGE_SEPARATE_DEBUG_HEADER)lpFileBase);
    }
    else if ( (pImgFileHdr->Machine == IMAGE_FILE_MACHINE_I386)  ||
              (pImgFileHdr->Machine == IMAGE_FILE_MACHINE_AMD64) ||
              (pImgFileHdr->Machine == IMAGE_FILE_MACHINE_ALPHA) ||
              (pImgFileHdr->Machine == IMAGE_FILE_MACHINE_IA64) )
    {
        if (pImgFileHdr->SizeOfOptionalHeader == 0)  // 0 optional header
            DumpObjFile (pImgFileHdr);               // means it's an OBJ

        else if (pImgFileHdr->SizeOfOptionalHeader == IMAGE_SIZEOF_ROM_OPTIONAL_HEADER)
        {
          DumpROMImage ((const IMAGE_ROM_HEADERS*)pImgFileHdr);
        }
    }
    else if (!strncmp((const char*)lpFileBase, IMAGE_ARCHIVE_START, IMAGE_ARCHIVE_START_SIZE))
    {
        DumpLibFile (lpFileBase);
    }
    else
        printf("unrecognized file format\n");

    UnmapViewOfFile (lpFileBase);
    CloseHandle (hFileMapping);
    CloseHandle (hFile);
}

/*
 * Process all the command line arguments.
 */
static void ProcessCommandLine (int argc, char *argv[])
{
  int opt;

  if (argc == 1)
  {
    printf (HelpText);
    exit (1);
  }

  /* Parse options.  */
  while ((opt = getopt (argc, argv, "?hABDHILPRSd")) != EOF)
  {
    switch (opt)
    {
      case '?':
      case 'h':
           printf (HelpText);
           exit (0);

      case 'A':
           fShowRelocations = TRUE;
           fShowRawSectionData = TRUE;
           fShowSymbolTable = TRUE;
           fShowLineNumbers = TRUE;
           fShowIATentries = TRUE;
           fShowPDATA = TRUE;
           fShowResources = TRUE;
           break;

      case 'B':
           fShowRelocations = TRUE;
           break;

      case 'D':
           fDemangle = FALSE;
           break;

      case 'H':
           fShowRawSectionData = TRUE;
           break;

      case 'I':
           fShowIATentries = TRUE;
           break;

      case 'L':
           fShowLineNumbers = TRUE;
           break;

      case 'P':
           fShowPDATA = TRUE;
           break;

      case 'R':
           fShowResources = TRUE;
           break;

      case 'S':
           fShowSymbolTable = TRUE;
           break;

      case 'd':
           fDebug++;
           break;
    }
  }
}

int main (int argc, char **argv)
{
  int i;

  program_name = argv[0];

  ProcessCommandLine (argc, argv);

#if 0
  printf ("optind: %d\n", optind);
  for (i = optind; i < argc; i++)
     printf ("argv[%d]: \"%s\"\n", i, argv[i]);
#endif

  init_dbghelp();
  init_imagehlp();
  sigsegv_handler_old = signal (SIGSEGV, sigsegv_handler);

  for (i = optind; i < argc; i++)
  {
    DumpFile (argv[i]);
    if (i < argc-1)
      puts ("==========================================="
            "===========================================");
  }
  return 0;
}


/* Return TRUE if all characters can occur in a mangled name.
 */
#define IS_DIGIT(c) ((c) >= '0' && (c) <= '9')
#define IS_ALPHA(c) (((c) >= 'a' && (c) <= 'z') || \
                     ((c) >= 'A' && (c) <= 'Z'))

#define IS_MANGLED_CHAR(c)  (IS_ALPHA(c) || IS_DIGIT(c) || \
                             (c) == '_' || (c) == '.' || (c) == '$')

static inline BOOL maybe_gxx_mangled_name (const char *name)
{
  if (name[0] == '?')
     return (FALSE);

  for (const char *p = name; *p; p++)
      if (!IS_MANGLED_CHAR(*p))
         return (FALSE);
  return (TRUE);
}

const char *undecorate_name (const char *name)
{
  static char und_name[1024];
  DWORD flags = UNDNAME_COMPLETE /* | UNDNAME_NAME_ONLY */;

  if (!fDemangle)
     return (name);

  und_name[0] = '\0';

  /* First try MS names. They start with a '?'.
   */
  if (p_UnDecorateSymbolName && name[0] == '?')
  {
    if ((*p_UnDecorateSymbolName)(name, und_name, sizeof(und_name), flags))
       return (und_name);
    if (fDebug)
    {
      snprintf (und_name, sizeof(und_name), "fail %lu on '%s'", GetLastError(), name);
      return (und_name);
    }
  }

#if 0
  fflush (stdout);

  if (setjmp(sigsegv_buf))
     return (name);
#endif

  /* Then attempt GNU demangling if the name contains legal mangling chars.
   * (unless it's a MS name starting with '?').
   */
  if (maybe_gxx_mangled_name(name))
  {
    int   options = DMGL_PARAMS | DMGL_ANSI | DMGL_TYPES /* | DMGL_RET_POSTFIX | DMGL_JAVA */ ;
    char *demangled = cplus_demangle_v3 (name, options);

    if (demangled)
       return (demangled); /* todo: free this */
  }

  /* Demangling failed; return 'name' unchanged.
   */
  return (name);
}


