//==================================
// PEDUMP - Matt Pietrek 1997-2001
// FILE: COMMON.CPP
//==================================

#include "Common.h"
#include "COFFSymbolTable.h"
#include "ExtrnVar.h"

struct PDB_INFO {
       CHAR   Signature[4];   // "NBxx"
       ULONG  Offset;         // always zero
       ULONG  sig;
       ULONG  age;
       CHAR   PdbName[_MAX_PATH];
     };


//
// Dump the COFF debug information header
//
void DumpCOFFHeader (const IMAGE_COFF_SYMBOLS_HEADER * pDbgInfo)
{
  printf ("COFF Debug Info Header\n");
  printf ("  NumberOfSymbols:      %08lX\n", pDbgInfo->NumberOfSymbols);
  printf ("  LvaToFirstSymbol:     %08lX\n", pDbgInfo->LvaToFirstSymbol);
  printf ("  NumberOfLinenumbers:  %08lX\n", pDbgInfo->NumberOfLinenumbers);
  printf ("  LvaToFirstLinenumber: %08lX\n", pDbgInfo->LvaToFirstLinenumber);
  printf ("  RvaToFirstByteOfCode: %08lX\n", pDbgInfo->RvaToFirstByteOfCode);
  printf ("  RvaToLastByteOfCode:  %08lX\n", pDbgInfo->RvaToLastByteOfCode);
  printf ("  RvaToFirstByteOfData: %08lX\n", pDbgInfo->RvaToFirstByteOfData);
  printf ("  RvaToLastByteOfData:  %08lX\n", pDbgInfo->RvaToLastByteOfData);
}

//
// Given a COFF symbol table index, look up its name.  This function assumes
// that the COFFSymbolCount and *COFFSymbolTable variables have been already
// set up.
//
static BOOL LookupSymbolName (DWORD index, char *buffer, UINT length)
{
  if (!g_pCOFFSymbolTable)
    return FALSE;

  class COFFSymbol *pSymbol = g_pCOFFSymbolTable->GetSymbolFromIndex (index);

  if (!pSymbol)
    return FALSE;

  lstrcpyn (buffer, pSymbol->GetName(), length);

  delete pSymbol;

  return TRUE;
}

//
// Dump a range of line numbers from the COFF debug information
//
void DumpLineNumbers (const IMAGE_LINENUMBER *pln, DWORD count)
{
  char  buffer[64];
  DWORD i;

  printf ("Line Numbers\n");

  for (i = 0; i < count; i++)
  {
    if (pln->Linenumber == 0)   // A symbol table index
    {
      buffer[0] = 0;
      LookupSymbolName (pln->Type.SymbolTableIndex, buffer, sizeof(buffer));
      printf ("SymIndex: %lX (%s)\n",
              pln->Type.SymbolTableIndex, buffer);
    }
    else                        // A regular line number
      printf (" Addr: %05lX  Line: %04u\n",
              pln->Type.VirtualAddress, pln->Linenumber);
    pln++;
  }
}

//
// Used by the DumpSymbolTable() routine.  It purpose is to filter out
// the non-normal section numbers and give them meaningful names.
//
static void GetSectionName (WORD section, char *buffer, unsigned cbBuffer)
{
  char tempbuffer[10];

  switch ((SHORT) section)
  {
    case IMAGE_SYM_UNDEFINED:
         strcpy (tempbuffer, "UNDEF");
         break;
    case IMAGE_SYM_ABSOLUTE:
         strcpy (tempbuffer, "ABS");
         break;
    case IMAGE_SYM_DEBUG:
         strcpy (tempbuffer, "DEBUG");
         break;
    default:
         sprintf (tempbuffer, "%X", section);
  }
  strncpy (buffer, tempbuffer, cbBuffer - 1);
}

//
// Dumps a COFF symbol table from an EXE or OBJ
//
void DumpCOFFSymbolTable (class COFFSymbolTable *pSymTab)
{
  printf ("Symbol Table - %X entries  (* = auxillary symbol)\n",
          pSymTab->GetNumberOfSymbols());

  printf ("Indx Sectn Value    Type  Storage  Name\n"
          "---- ----- -------- ----- -------  --------\n");

  class COFFSymbol *pSymbol = pSymTab->GetNextSymbol (0);

  while (pSymbol)
  {
    char szSection[10];

    GetSectionName (pSymbol->GetSectionNumber(), szSection, sizeof(szSection));

    printf ("%04lX %5s %08lX  %s %-8s %s\n",
            pSymbol->GetIndex(), szSection, pSymbol->GetValue(),
            pSymbol->GetTypeName(), pSymbol->GetStorageClassName(),
            undecorate_name (pSymbol->GetName()));

    if (pSymbol->GetNumberOfAuxSymbols())
    {
      char szAuxSymbol[1024];

      if (pSymbol->GetAuxSymbolAsString (szAuxSymbol, sizeof(szAuxSymbol)))
         printf ("     * %s\n", szAuxSymbol);
    }
    pSymbol = pSymTab->GetNextSymbol (pSymbol);
  }
}

void DumpMiscDebugInfo (const IMAGE_DEBUG_MISC * pMiscDebugInfo)
{
  if (IMAGE_DEBUG_MISC_EXENAME != pMiscDebugInfo->DataType)
  {
    printf ("Unknown Miscellaneous Debug Information type: %lu\n", pMiscDebugInfo->DataType);
    return;
  }

  printf ("Miscellaneous Debug Information\n");
  printf ("  %s\n", pMiscDebugInfo->Data);
}

void DumpCVDebugInfoRecord (const DWORD * pCVHeader)
{
  const struct PDB_INFO *pPDBInfo;

  printf ("CodeView Signature: %08lX\n", *pCVHeader);

  if ('SDSR' == (UINT32) * pCVHeader)
  {
    printf ("New format Visual Studio .NET (or later) debug information\n");
    return;
  }

  if ('BN' != (unsigned short) *pCVHeader)
  {
    printf ("Unhandled CodeView Information format %.4s\n", (const char *) pCVHeader);
    return;
  }

  if ('01BN' == (unsigned short) *pCVHeader) // NB10 == PDB file info
  {
    pPDBInfo = (const struct PDB_INFO *) pCVHeader;

    printf ("  Offset: %08lX  Signature: %08lX  Age: %08lX\n", pPDBInfo->Offset, pPDBInfo->sig, pPDBInfo->age);
    printf ("  File: %s\n", pPDBInfo->PdbName);
  }
  else                          // Should be NB09 or NB11
  {
    DumpCVSymbolTable ((const BYTE*)pCVHeader);
  }
}
